/*
 * Jalview - A Sequence Alignment Editor and Viewer
 * Copyright (C) 2006 AM Waterhouse, J Procter, G Barton, M Clamp, S Searle
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 */

package jalview.appletgui;

import java.util.*;

import java.awt.*;

import jalview.analysis.*;
import jalview.bin.*;
import jalview.datamodel.*;
import jalview.schemes.*;

public class AlignViewport
{
  int startRes;
  int endRes;

  int startSeq;
  int endSeq;


  boolean cursorMode = false;

  boolean showJVSuffix = true;
  boolean showText = true;
  boolean showColourText = false;
  boolean showBoxes = true;
  boolean wrapAlignment = false;
  boolean renderGaps = true;
  boolean showSequenceFeatures = false;
  boolean showAnnotation = true;
  boolean showConservation = true;
  boolean showQuality = true;
  boolean showConsensus = true;

  boolean colourAppliesToAllGroups = true;
  ColourSchemeI globalColourScheme = null;
  boolean conservationColourSelected = false;
  boolean abovePIDThreshold = false;

  SequenceGroup selectionGroup;

  int charHeight;
  int charWidth;
  int wrappedWidth;

  Font font = new Font("SansSerif", Font.PLAIN, 10);
  boolean validCharWidth = true;
  AlignmentI alignment;

  ColumnSelection colSel = new ColumnSelection();

  int threshold;
  int increment;

  NJTree currentTree = null;

  boolean scaleAboveWrapped = true;
  boolean scaleLeftWrapped = true;
  boolean scaleRightWrapped = true;

  // The following vector holds the features which are
 // currently visible, in the correct order or rendering
  Hashtable featuresDisplayed;

  boolean hasHiddenColumns = false;
  boolean hasHiddenRows = false;
  boolean showHiddenMarkers = true;


  public Vector vconsensus;
  AlignmentAnnotation consensus;
  AlignmentAnnotation conservation;
  AlignmentAnnotation quality;

  boolean autocalculateConsensus = true;

  public int ConsPercGaps = 25; // JBPNote : This should be a scalable property!

  private java.beans.PropertyChangeSupport changeSupport = new java.beans.PropertyChangeSupport(this);

  boolean ignoreGapsInConsensusCalculation = false;

  jalview.bin.JalviewLite applet;

  boolean MAC = false;

  public AlignViewport(AlignmentI al, JalviewLite applet)
  {
    this.applet = applet;
    setAlignment(al);
    this.startRes = 0;
    this.endRes = al.getWidth() - 1;
    this.startSeq = 0;
    this.endSeq = al.getHeight() - 1;
    setFont(font);

    if(System.getProperty("os.name").startsWith("Mac"))
      MAC = true;

    if (applet != null)
    {
      String param = applet.getParameter("showFullId");
      if (param != null)
      {
        showJVSuffix = Boolean.valueOf(param).booleanValue();
      }

      param = applet.getParameter("showAnnotation");
      if (param != null)
      {
        showAnnotation = Boolean.valueOf(param).booleanValue();
      }

      param = applet.getParameter("showConservation");
      if (param != null)
      {
        showConservation = Boolean.valueOf(param).booleanValue();
      }

      param = applet.getParameter("showQuality");
      if (param != null)
      {
        showQuality = Boolean.valueOf(param).booleanValue();
      }

      param = applet.getParameter("showConsensus");
      if (param != null)
      {
        showConsensus = Boolean.valueOf(param).booleanValue();
      }
    }
    // We must set conservation and consensus before setting colour,
    // as Blosum and Clustal require this to be done
    updateConservation();
    updateConsensus();


    if (applet != null)
    {
      String colour = applet.getParameter("defaultColour");

      if(colour == null)
      {
        colour = applet.getParameter("userDefinedColour");
        if(colour !=null)
          colour = "User Defined";
      }

      if(colour != null)
      {
        globalColourScheme = ColourSchemeProperty.getColour(alignment, colour);
        if (globalColourScheme != null)
        {
          globalColourScheme.setConsensus(vconsensus);
        }
      }

      if(applet.getParameter("userDefinedColour")!=null)
      {
        ((UserColourScheme)globalColourScheme).parseAppletParameter(
            applet.getParameter("userDefinedColour"));
      }


    }
  }

  public void showSequenceFeatures(boolean b)
  {
    showSequenceFeatures = b;
  }

  public boolean getShowSequenceFeatures()
  {
    return showSequenceFeatures;
  }


  public void updateConservation()
  {
    if(alignment.isNucleotide())
          return;

    Conservation cons = new jalview.analysis.Conservation("All",
        jalview.schemes.ResidueProperties.propHash, 3,
        alignment.getSequences(), 0,
        alignment.getWidth() - 1);
    cons.calculate();
    cons.verdict(false, ConsPercGaps);
    cons.findQuality();
    int alWidth = alignment.getWidth();
    Annotation[] annotations = new Annotation[alWidth];
    Annotation[] qannotations = new Annotation[alWidth];
    String sequence = cons.getConsSequence().getSequence();
    float minR, minG, minB, maxR, maxG, maxB;
    minR = 0.3f;
    minG = 0.0f;
    minB = 0f;
    maxR = 1.0f - minR;
    maxG = 0.9f - minG;
    maxB = 0f - minB; // scalable range for colouring both Conservation and Quality
    float min = 0f;
    float max = 11f;
    float qmin = cons.qualityRange[0].floatValue();
    float qmax = cons.qualityRange[1].floatValue();

    for (int i = 0; i < alWidth; i++)
    {
      float value = 0;
      try
      {
        value = Integer.parseInt(sequence.charAt(i) + "");
      }
      catch (Exception ex)
      {
        if (sequence.charAt(i) == '*')
        {
          value = 11;
        }
        if (sequence.charAt(i) == '+')
        {
          value = 10;
        }
      }
      float vprop = value - min;
      vprop /= max;

      annotations[i] = new Annotation(sequence.charAt(i) + "",
                                      "", ' ', value,
                                      new Color(minR + maxR * vprop,
                                                minG + maxG * vprop,
                                                minB + maxB * vprop));
      // Quality calc
      value = ( (Double) cons.quality.elementAt(i)).floatValue();
      vprop = value - qmin;
      vprop /= qmax;
      qannotations[i] = new Annotation(" ",
                                       String.valueOf(value), ' ', value,
                                       new
                                       Color(minR + maxR * vprop,
                                             minG + maxG * vprop,
                                             minB + maxB * vprop));
    }

    if (conservation == null)
    {
      conservation = new AlignmentAnnotation("Conservation",
                                             "Conservation of total alignment less than " +
                                             ConsPercGaps + "% gaps",
                                             annotations,
                                             0f, // cons.qualityRange[0].floatValue(),
                                             11f, // cons.qualityRange[1].floatValue()
                                             AlignmentAnnotation.BAR_GRAPH);
      if (showConservation)
      {
        alignment.addAnnotation(conservation);
      }
      quality = new AlignmentAnnotation("Quality",
                                        "Alignment Quality based on Blosum62 scores",
                                        qannotations,
                                        cons.qualityRange[0].floatValue(),
                                        cons.qualityRange[1].floatValue(),
                                        AlignmentAnnotation.BAR_GRAPH);
      if (showQuality)
      {
        alignment.addAnnotation(quality);
      }
    }
    else
    {
      conservation.annotations = annotations;
      quality.annotations = qannotations;
      quality.graphMax = cons.qualityRange[1].floatValue();
    }

  }

  public void updateConsensus()
  {
    Annotation[] annotations = new Annotation[alignment.getWidth()];

    // this routine prevents vconsensus becoming a new object each time
    // consenus is calculated. Important for speed of Blosum62
    // and PID colouring of alignment
    if (vconsensus == null)
    {
      vconsensus = alignment.getAAFrequency();
    }
    else
    {
      Vector temp = alignment.getAAFrequency();
      vconsensus.removeAllElements();
      Enumeration e = temp.elements();
      while (e.hasMoreElements())
      {
        vconsensus.addElement(e.nextElement());
      }
    }
    Hashtable hash = null;
    for (int i = 0; i < alignment.getWidth(); i++)
    {
      hash = (Hashtable) vconsensus.elementAt(i);
      float value = 0;
      if(ignoreGapsInConsensusCalculation)
        value = ((Float)hash.get("pid_nogaps")).floatValue();
      else
        value = ((Float)hash.get("pid_gaps")).floatValue();

      String maxRes = hash.get("maxResidue").toString();
      String mouseOver = hash.get("maxResidue") + " ";
      if (maxRes.length() > 1)
      {
        mouseOver = "[" + maxRes + "] ";
        maxRes = "+";
      }


      mouseOver += (int) value + "%";
      annotations[i] = new Annotation(maxRes, mouseOver, ' ', value);

    }

    if (consensus == null)
    {
      consensus = new AlignmentAnnotation("Consensus",
                                          "PID", annotations, 0f, 100f, AlignmentAnnotation.BAR_GRAPH);
      if (showConsensus)
      {
        alignment.addAnnotation(consensus);
      }
    }
    else
    {
      consensus.annotations = annotations;
    }

    if(globalColourScheme!=null)
          globalColourScheme.setConsensus(vconsensus);

  }
  /**
   * get the consensus sequence as displayed under the PID consensus annotation row.
   * @return consensus sequence as a new sequence object
   */
  /**
   * get the consensus sequence as displayed under the PID consensus annotation row.
   * @return consensus sequence as a new sequence object
   */
  public SequenceI getConsensusSeq() {
    if (consensus==null)
      updateConsensus();
    if (consensus==null)
      return null;
    StringBuffer seqs=new StringBuffer();
    for (int i=0; i<consensus.annotations.length; i++) {
      if (consensus.annotations[i]!=null) {
        if (consensus.annotations[i].description.charAt(0) == '[')
          seqs.append(consensus.annotations[i].description.charAt(1));          
        else
          seqs.append(consensus.annotations[i].displayCharacter);          
      }
    }
    SequenceI sq = new Sequence("Consensus", seqs.toString());
    sq.setDescription("Percentage Identity Consensus "+((ignoreGapsInConsensusCalculation) ? " without gaps" : ""));
    return sq;
  }
  public SequenceGroup getSelectionGroup()
  {
    return selectionGroup;
  }

  public void setSelectionGroup(SequenceGroup sg)
  {
    selectionGroup = sg;
  }

  public boolean getConservationSelected()
  {
    return conservationColourSelected;
  }

  public void setConservationSelected(boolean b)
  {
    conservationColourSelected = b;
  }

  public boolean getAbovePIDThreshold()
  {
    return abovePIDThreshold;
  }

  public void setAbovePIDThreshold(boolean b)
  {
    abovePIDThreshold = b;
  }

  public int getStartRes()
  {
    return startRes;
  }

  public int getEndRes()
  {
    return endRes;
  }

  public int getStartSeq()
  {
    return startSeq;
  }

  public void setGlobalColourScheme(ColourSchemeI cs)
  {
    globalColourScheme = cs;
  }

  public ColourSchemeI getGlobalColourScheme()
  {
    return globalColourScheme;
  }

  public void setStartRes(int res)
  {
    this.startRes = res;
  }

  public void setStartSeq(int seq)
  {
    this.startSeq = seq;
  }

  public void setEndRes(int res)
  {
    if (res > alignment.getWidth() - 1)
    {
      // log.System.out.println(" Corrected res from " + res + " to maximum " + (alignment.getWidth()-1));
      res = alignment.getWidth() - 1;
    }
    if (res < 0)
    {
      res = 0;
    }
    this.endRes = res;
  }

  public void setEndSeq(int seq)
  {
    if (seq > alignment.getHeight())
    {
      seq = alignment.getHeight();
    }
    if (seq < 0)
    {
      seq = 0;
    }
    this.endSeq = seq;
  }

  public int getEndSeq()
  {
    return endSeq;
  }

  java.awt.Frame nullFrame;
  public void setFont(Font f)
  {
    font = f;
    if(nullFrame == null)
    {
      nullFrame = new java.awt.Frame();
      nullFrame.addNotify();
    }

    java.awt.FontMetrics fm = nullFrame.getGraphics().getFontMetrics(font);
    setCharHeight(fm.getHeight());
    charWidth = fm.charWidth('M');
  }

  public Font getFont()
  {
    return font;
  }

  public int getCharWidth()
  {
    return charWidth;
  }

  public void setCharHeight(int h)
  {
    this.charHeight = h;
  }

  public int getCharHeight()
  {
    return charHeight;
  }

  public void setWrappedWidth(int w)
  {
    this.wrappedWidth = w;
  }

  public int getwrappedWidth()
  {
    return wrappedWidth;
  }

  public AlignmentI getAlignment()
  {
    return alignment;
  }

  public void setAlignment(AlignmentI align)
  {
    this.alignment = align;
  }

  public void setWrapAlignment(boolean state)
  {
    wrapAlignment = state;
  }

  public void setShowText(boolean state)
  {
    showText = state;
  }

  public void setRenderGaps(boolean state)
  {
    renderGaps = state;
  }

  public boolean getColourText()
  {
    return showColourText;
  }

  public void setColourText(boolean state)
  {
    showColourText = state;
  }

  public void setShowBoxes(boolean state)
  {
    showBoxes = state;
  }

  public boolean getWrapAlignment()
  {
    return wrapAlignment;
  }

  public boolean getShowText()
  {
    return showText;
  }

  public boolean getShowBoxes()
  {
    return showBoxes;
  }

  public char getGapCharacter()
  {
    return getAlignment().getGapCharacter();
  }

  public void setGapCharacter(char gap)
  {
    if (getAlignment() != null)
    {
      getAlignment().setGapCharacter(gap);
    }
  }

  public void setThreshold(int thresh)
  {
    threshold = thresh;
  }

  public int getThreshold()
  {
    return threshold;
  }

  public void setIncrement(int inc)
  {
    increment = inc;
  }

  public int getIncrement()
  {
    return increment;
  }

  public void setHiddenColumns(ColumnSelection colsel)
  {
    this.colSel = colsel;
    if(colSel.getHiddenColumns()!=null)
      hasHiddenColumns = true;
  }

  public ColumnSelection getColumnSelection()
  {
    return colSel;
  }

  public void resetSeqLimits(int height)
  {
    setEndSeq(height / getCharHeight());
  }

  public void setCurrentTree(NJTree tree)
  {
    currentTree = tree;
  }

  public NJTree getCurrentTree()
  {
    return currentTree;
  }

  public void setColourAppliesToAllGroups(boolean b)
  {
    colourAppliesToAllGroups = b;
  }

  public boolean getColourAppliesToAllGroups()
  {
    return colourAppliesToAllGroups;
  }

  public boolean getShowJVSuffix()
  {
    return showJVSuffix;
  }

  public void setShowJVSuffix(boolean b)
  {
    showJVSuffix = b;
  }

  public boolean getShowAnnotation()
  {
    return showAnnotation;
  }

  public void setShowAnnotation(boolean b)
  {
    showAnnotation = b;
  }

  public boolean getScaleAboveWrapped()
  {
    return scaleAboveWrapped;
  }

  public boolean getScaleLeftWrapped()
  {
    return scaleLeftWrapped;
  }

  public boolean getScaleRightWrapped()
  {
    return scaleRightWrapped;
  }

  public void setScaleAboveWrapped(boolean b)
  {
    scaleAboveWrapped = b;
  }

  public void setScaleLeftWrapped(boolean b)
  {
    scaleLeftWrapped = b;
  }

  public void setScaleRightWrapped(boolean b)
  {
    scaleRightWrapped = b;
  }

  public void setIgnoreGapsConsensus(boolean b)
  {
    ignoreGapsInConsensusCalculation = b;
    updateConsensus();
    if (globalColourScheme!=null)
    {
      globalColourScheme.setThreshold(globalColourScheme.getThreshold(),
          ignoreGapsInConsensusCalculation);

    }
  }

  /**
   * Property change listener for changes in alignment
   *
   * @param listener DOCUMENT ME!
   */
  public void addPropertyChangeListener(
      java.beans.PropertyChangeListener listener)
  {
      changeSupport.addPropertyChangeListener(listener);
  }

  /**
   * DOCUMENT ME!
   *
   * @param listener DOCUMENT ME!
   */
  public void removePropertyChangeListener(
      java.beans.PropertyChangeListener listener)
  {
      changeSupport.removePropertyChangeListener(listener);
  }

  /**
   * Property change listener for changes in alignment
   *
   * @param prop DOCUMENT ME!
   * @param oldvalue DOCUMENT ME!
   * @param newvalue DOCUMENT ME!
   */
  public void firePropertyChange(String prop, Object oldvalue, Object newvalue)
  {
      changeSupport.firePropertyChange(prop, oldvalue, newvalue);
  }



  public boolean getIgnoreGapsConsensus()
  {
    return ignoreGapsInConsensusCalculation;
  }
  public void hideSelectedColumns()
  {
    if (colSel.size() < 1)
      return;

    colSel.hideSelectedColumns();
    setSelectionGroup(null);

    hasHiddenColumns = true;
  }

  public void invertColumnSelection()
  {
    int column;
    for (int i = 0; i < alignment.getWidth(); i++)
    {
      column = i;

      if (colSel.contains(column))
        colSel.removeElement(column);
      else
        colSel.addElement(column);

    }
  }


  public void hideColumns(int start, int end)
  {
    if(start==end)
      colSel.hideColumns(start);
    else
      colSel.hideColumns(start, end);

    hasHiddenColumns = true;
  }

  public void hideSequence(SequenceI seq)
  {
    if(seq!=null)
    {
      alignment.getHiddenSequences().hideSequence(seq);
      hasHiddenRows = true;
      firePropertyChange("alignment", null, alignment.getSequences());
    }
  }

  public void hideAllSelectedSeqs()
  {
    if (selectionGroup == null)
      return;

    SequenceI[] seqs = selectionGroup.getSequencesInOrder(alignment);

    for (int i = 0; i < seqs.length; i++)
    {
      alignment.getHiddenSequences().hideSequence(seqs[i]);
    }
    firePropertyChange("alignment", null, alignment.getSequences());
    hasHiddenRows = true;
    setSelectionGroup(null);
  }

  public void showColumn(int col)
  {
    colSel.revealHiddenColumns(col);
    if(colSel.getHiddenColumns()==null)
      hasHiddenColumns = false;
  }

  public void showAllHiddenColumns()
  {
    colSel.revealAllHiddenColumns();
    hasHiddenColumns = false;
  }

  public void showAllHiddenSeqs()
  {
    if(alignment.getHiddenSequences().getSize()>0)
    {
      if(selectionGroup==null)
      {
        selectionGroup = new SequenceGroup();
        selectionGroup.setEndRes(alignment.getWidth()-1);
      }
      Vector tmp = alignment.getHiddenSequences().showAll();
      for(int t=0; t<tmp.size(); t++)
      {
        selectionGroup.addSequence(
            (SequenceI)tmp.elementAt(t), false
            );
      }
      firePropertyChange("alignment", null, alignment.getSequences());
      hasHiddenRows = false;
    }
  }

  public int adjustForHiddenSeqs(int alignmentIndex)
  {
    return alignment.getHiddenSequences().adjustForHiddenSeqs(alignmentIndex);
  }

  /**
   * This method returns the a new SequenceI [] with
   * the selection sequence and start and end points adjusted
   * @return String[]
   */
  public SequenceI[] getSelectionAsNewSequence()
  {
    SequenceI[] sequences;

    if (selectionGroup == null)
      sequences = alignment.getSequencesArray();
    else
      sequences = selectionGroup.getSelectionAsNewSequences(alignment);

    return sequences;
  }

  /**
   * This method returns the visible alignment as text, as
   * seen on the GUI, ie if columns are hidden they will not
   * be returned in the result.
   * Use this for calculating trees, PCA, redundancy etc on views
   * which contain hidden columns.
   * @return String[]
   */
  public jalview.datamodel.CigarArray getViewAsCigars(boolean selectedRegionOnly)
  {
    CigarArray selection=null;
    SequenceI [] seqs= null;
    int i, iSize;
    int start = 0, end = 0;
    if(selectedRegionOnly && selectionGroup!=null)
    {
      iSize = selectionGroup.getSize(false);
      seqs = selectionGroup.getSequencesInOrder(alignment);
      start = selectionGroup.getStartRes();
      end = selectionGroup.getEndRes(); // inclusive for start and end in SeqCigar constructor
    }
    else
    {
      iSize = alignment.getHeight();
      seqs = alignment.getSequencesArray();
      end = alignment.getWidth()-1;
    }
    SeqCigar[] selseqs = new SeqCigar[iSize];
    for(i=0; i<iSize; i++)
    {
      selseqs[i] = new SeqCigar(seqs[i], start, end);
    }
    selection=new CigarArray(selseqs);
    // now construct the CigarArray operations
    if (hasHiddenColumns) {
      Vector regions = colSel.getHiddenColumns();
      int [] region;
      int hideStart, hideEnd;
      int last=start;
      for (int j = 0; last<end & j < regions.size(); j++)
      {
        region = (int[]) regions.elementAt(j);
        hideStart = region[0];
        hideEnd = region[1];
        // edit hidden regions to selection range
        if(hideStart<last) {
          if (hideEnd > last)
          {
            hideStart = last;
          } else
            continue;
        }

        if (hideStart>end)
          break;

        if (hideEnd>end)
          hideEnd=end;

        if (hideStart>hideEnd)
          break;
        /**
         * form operations...
         */
        if (last<hideStart)
          selection.addOperation(CigarArray.M, hideStart-last);
        selection.addOperation(CigarArray.D, 1+hideEnd-hideStart);
        last = hideEnd+1;
      }
      // Final match if necessary.
      if (last<end)
        selection.addOperation(CigarArray.M, end-last);
    } else {
      selection.addOperation(CigarArray.M, end-start);
    }
    return selection;
  }
  /**
   * return a compact representation of the current alignment selection to
   * pass to an analysis function
   * @param selectedOnly boolean true to just return the selected view
   * @return AlignmentView
   */
  jalview.datamodel.AlignmentView getAlignmentView(boolean selectedOnly) {
    // JBPNote:
    // this is here because the AlignmentView constructor modifies the CigarArray
    // object. Refactoring of Cigar and alignment view representation should
    // be done to remove redundancy.
    CigarArray aligview = getViewAsCigars(selectedOnly);
    if (aligview!=null)
      return new AlignmentView(aligview);
    return null;
  }
  /**
   * This method returns the visible alignment as text, as
   * seen on the GUI, ie if columns are hidden they will not
   * be returned in the result.
   * Use this for calculating trees, PCA, redundancy etc on views
   * which contain hidden columns.
   * @return String[]
   */
  public String [] getViewAsString(boolean selectedRegionOnly)
  {
    String [] selection = null;
    SequenceI [] seqs= null;
    int i, iSize;
    int start = 0, end = 0;
    if(selectedRegionOnly && selectionGroup!=null)
    {
      iSize = selectionGroup.getSize(false);
      seqs = selectionGroup.getSequencesInOrder(alignment);
      start = selectionGroup.getStartRes();
      end = selectionGroup.getEndRes()+1;
    }
    else
    {
      iSize = alignment.getHeight();
      seqs = alignment.getSequencesArray();
      end = alignment.getWidth();
    }

    selection = new String[iSize];


    for(i=0; i<iSize; i++)
    {
      if (hasHiddenColumns)
      {
           StringBuffer visibleSeq = new StringBuffer();
           Vector regions = colSel.getHiddenColumns();

           int blockStart = start, blockEnd=end;
           int [] region;
           int hideStart, hideEnd;

           for (int j = 0; j < regions.size(); j++)
           {
             region = (int[]) regions.elementAt(j);
             hideStart = region[0];
             hideEnd = region[1];

             if(hideStart < start)
             {
               continue;
             }

             blockStart = Math.min(blockStart, hideEnd+1);
             blockEnd = Math.min(blockEnd, hideStart);

             if(blockStart>blockEnd)
             {
                break;
             }


             visibleSeq.append(seqs[i].getSequence(blockStart, blockEnd));

             blockStart = hideEnd+1;
             blockEnd = end;
           }

           if(end>blockStart)
             visibleSeq.append(seqs[i].getSequence(blockStart, end));

           selection[i] = visibleSeq.toString();
      }
      else
      {
        selection[i] = seqs[i].getSequence(start, end);
      }
    }

    return selection;
  }

  public boolean getShowHiddenMarkers()
  {
    return showHiddenMarkers;
  }

  public void setShowHiddenMarkers(boolean show)
  {
    showHiddenMarkers = show;
  }


}
